@@ -1,11 +1,10 @@
 use core::ffi::c_void;
-use std::boxed::Box;
 use std::fmt::Debug;
 use std::fmt::Formatter;
 use std::fmt::Result as FmtResult;
 use std::ops::Deref as _;
+use std::ops::DerefMut as _;
 use std::os::raw::c_ulong;
-use std::os::unix::io::AsFd;
 use std::os::unix::prelude::AsRawFd;
 use std::os::unix::prelude::BorrowedFd;
 use std::ptr::null_mut;
@@ -13,11 +12,12 @@ use std::ptr::NonNull;
 use std::slice;
 use std::time::Duration;
 
-use crate::libbpf_sys;
 use crate::util;
+use crate::util::validate_bpf_ret;
 use crate::AsRawLibbpf;
 use crate::Error;
-use crate::MapHandle;
+use crate::ErrorExt as _;
+use crate::MapCore;
 use crate::MapType;
 use crate::Result;
 
@@ -67,11 +67,12 @@ impl<'slf, 'cb: 'slf> RingBufferBuilder<'slf, 'cb> {
     /// manager. The callback should take one argument, a slice of raw bytes,
     /// and return an i32.
     ///
-    /// Non-zero return values in the callback will stop ring buffer consumption early.
+    /// Negative return values in the callback will stop ring buffer consumption early and
+    /// propagate the error code to the polling caller.
     ///
     /// The callback provides a raw byte slice. You may find libraries such as
     /// [`plain`](https://crates.io/crates/plain) helpful.
-    pub fn add<NewF>(&mut self, map: &'slf MapHandle, callback: NewF) -> Result<&mut Self>
+    pub fn add<NewF>(&mut self, map: &'slf dyn MapCore, callback: NewF) -> Result<&mut Self>
     where
         NewF: FnMut(&[u8]) -> i32 + 'cb,
     {
@@ -86,45 +87,55 @@ impl<'slf, 'cb: 'slf> RingBufferBuilder<'slf, 'cb> {
     /// Build a new [`RingBuffer`]. Must have added at least one ringbuf.
     pub fn build(self) -> Result<RingBuffer<'cb>> {
         let mut cbs = vec![];
-        let mut ptr: Option<NonNull<libbpf_sys::ring_buffer>> = None;
+        let mut rb_ptr: Option<NonNull<libbpf_sys::ring_buffer>> = None;
         let c_sample_cb: libbpf_sys::ring_buffer_sample_fn = Some(Self::call_sample_cb);
 
         for (fd, callback) in self.fd_callbacks {
-            let sample_cb_ptr = Box::into_raw(Box::new(callback));
-            match ptr {
+            let mut sample_cb = Box::new(callback);
+            match rb_ptr {
                 None => {
                     // Allocate a new ringbuf manager and add a ringbuf to it
-                    ptr = Some(util::create_bpf_entity_checked(|| unsafe {
+                    // SAFETY: All pointers are valid or rightly NULL.
+                    //         The object referenced by `sample_cb` is
+                    //         not modified by `libbpf`
+                    let ptr = unsafe {
                         libbpf_sys::ring_buffer__new(
                             fd.as_raw_fd(),
                             c_sample_cb,
-                            sample_cb_ptr as *mut _,
+                            sample_cb.deref_mut() as *mut _ as *mut _,
                             null_mut(),
                         )
-                    })?);
+                    };
+                    let ptr = validate_bpf_ret(ptr).context("failed to create new ring buffer")?;
+                    rb_ptr = Some(ptr)
                 }
-                Some(ptr) => {
+                Some(mut ptr) => {
                     // Add a ringbuf to the existing ringbuf manager
+                    // SAFETY: All pointers are valid or rightly NULL.
+                    //         The object referenced by `sample_cb` is
+                    //         not modified by `libbpf`
                     let err = unsafe {
                         libbpf_sys::ring_buffer__add(
                             ptr.as_ptr(),
                             fd.as_raw_fd(),
                             c_sample_cb,
-                            sample_cb_ptr as *mut _,
+                            sample_cb.deref_mut() as *mut _ as *mut _,
                         )
                     };
 
                     // Handle errors
                     if err != 0 {
+                        // SAFETY: The pointer is valid.
+                        let () = unsafe { libbpf_sys::ring_buffer__free(ptr.as_mut()) };
                         return Err(Error::from_raw_os_error(err));
                     }
                 }
             }
 
-            unsafe { cbs.push(Box::from_raw(sample_cb_ptr)) };
+            let () = cbs.push(sample_cb);
         }
 
-        match ptr {
+        match rb_ptr {
             Some(ptr) => Ok(RingBuffer { ptr, _cbs: cbs }),
             None => Err(Error::with_invalid_data(
                 "You must add at least one ring buffer map and callback before building",
@@ -156,24 +167,52 @@ pub struct RingBuffer<'cb> {
 impl RingBuffer<'_> {
     /// Poll from all open ring buffers, calling the registered callback for
     /// each one. Polls continually until we either run out of events to consume
-    /// or `timeout` is reached. If `timeout` is Duration::MAX, this will block
+    /// or `timeout` is reached. If `timeout` is `Duration::MAX`, this will block
     /// indefinitely until an event occurs.
-    pub fn poll(&self, timeout: Duration) -> Result<()> {
+    ///
+    /// Return the amount of events consumed, or a negative value in case of error.
+    pub fn poll_raw(&self, timeout: Duration) -> i32 {
         let mut timeout_ms = -1;
         if timeout != Duration::MAX {
             timeout_ms = timeout.as_millis() as i32;
         }
 
-        let ret = unsafe { libbpf_sys::ring_buffer__poll(self.ptr.as_ptr(), timeout_ms) };
+        unsafe { libbpf_sys::ring_buffer__poll(self.ptr.as_ptr(), timeout_ms) }
+    }
+
+    /// Poll from all open ring buffers, calling the registered callback for
+    /// each one. Polls continually until we either run out of events to consume
+    /// or `timeout` is reached. If `timeout` is `Duration::MAX`, this will block
+    /// indefinitely until an event occurs.
+    pub fn poll(&self, timeout: Duration) -> Result<()> {
+        let ret = self.poll_raw(timeout);
 
         util::parse_ret(ret)
     }
 
+    /// Greedily consume from all open ring buffers, calling the registered
+    /// callback for each one. Consumes continually until we run out of events
+    /// to consume or one of the callbacks returns a non-zero integer.
+    ///
+    /// Return the amount of events consumed, or a negative value in case of error.
+    pub fn consume_raw(&self) -> i32 {
+        unsafe { libbpf_sys::ring_buffer__consume(self.ptr.as_ptr()) }
+    }
+
+    /// Greedily consume from all open ring buffers, calling the registered
+    /// callback for each one. Continues until `len` items have been consumed,
+    /// no more events are available, or a callback returns a non-zero value.
+    ///
+    /// Return the amount of events consumed, or a negative value in case of error.
+    pub fn consume_raw_n(&self, len: usize) -> i32 {
+        unsafe { libbpf_sys::ring_buffer__consume_n(self.ptr.as_ptr(), len as libbpf_sys::size_t) }
+    }
+
     /// Greedily consume from all open ring buffers, calling the registered
     /// callback for each one. Consumes continually until we run out of events
     /// to consume or one of the callbacks returns a non-zero integer.
     pub fn consume(&self) -> Result<()> {
-        let ret = unsafe { libbpf_sys::ring_buffer__consume(self.ptr.as_ptr()) };
+        let ret = self.consume_raw();
 
         util::parse_ret(ret)
     }
